#include"McSimple.hpp"

McResult mc_nonvec(std::function<double(double)> f, RNorm& rnorm, int M){
    double mu = 0.0;
    double mu2 = 0.0;
    for(int i=0; i<M; ++i){
        double x = f(rnorm());
        mu += x;
        mu2 += x*x;
    }

    mu = mu / M;
    double stat = 1.96 * std::sqrt((mu2 / M - mu * mu) / M);
    return {mu, stat};
}

McResult mc_scalar(std::function<double(double)> f, RNorm& rnorm, int M){
    auto fgen = [&f, &rnorm](){ return f(rnorm());};
    Vector x(M);
    std::generate(x.begin(), x.end(), fgen);
    double mu = std::accumulate(x.begin(), x.end(), 0.0) / M;
    double mu2 = std::inner_product(x.begin(), x.end(), x.begin(), 0.0) / M;
    double stat = 1.96 * std::sqrt((mu2 - mu*mu) / M);
    return {mu, stat};
}

McResult mc_max_naive(RNorm& rnorm, int M, int N){
    double mu = 0.0;
    double mu2 = 0.0;
    for(int m = 0; m<M; ++m){
        double rmax = 0.0;
        for(int n=0; n<N; ++n)
            rmax = std::max(rmax, rnorm());
        mu += rmax;
        mu2 += rmax * rmax;
    }
    mu = mu / M;
    double stat = 1.96 * std::sqrt((mu2 / M - mu * mu) / M);
    return {mu, stat};
}

McResult mc_max_partial(RNorm& rnorm, int M, int N){
    double mu = 0.0;
    double mu2 = 0.0;
    Vector x(N);
    for(int m = 0; m<M; ++m){
        // I need this because using just rnorm instead of the lambda would produce the same
        // sample M times....
        std::ranges::generate(x, [&rnorm](){return rnorm();});
        double rmax = *std::max_element(x.begin(), x.end());
        mu += rmax;
        mu2 += rmax * rmax;
    }
    mu = mu / M;
    double stat = 1.96 * std::sqrt((mu2 / M - mu * mu) / M);
    return {mu, stat};
}

McResult mc_max(RNorm& rnorm, int M, int N){
    double mu = 0.0;
    double mu2 = 0.0;
    Vector x(N*M);
    std::generate(x.begin(), x.end(), rnorm);
    for(int m=0; m<M; ++m){
        double rmax = *std::max_element(x.begin() + m*N, x.begin() + (m+1)*N);
        mu += rmax;
        mu2 += rmax * rmax;
    }
    mu = mu / M;
    double stat = 1.96 * std::sqrt((mu2 / M - mu * mu) / M);
    return {mu, stat};
}