// run the MC algorithms
#include<iostream>
#include<chrono>
#include<cmath>
#include<algorithm>

#include "RNorm.hpp"
#include "McSimple.hpp"

using std::cout;
using Clock = std::chrono::steady_clock;
using std::chrono::duration_cast;
using std::chrono::milliseconds;

int main(){
    // generate our RNG
    const uint64_t seed = 12345678;
    RNorm rnorm(seed);
    const int M = 100000;
    const int N = 100;
    const int repN = 10; // number of repetitions for timing

    // run the first example.

    // first a warm-up before timing!
    auto start = Clock::now();
    McResult mc = mc_nonvec([](double x) {return std::sin(x);}, rnorm, M);
    auto end = Clock::now();
	auto diff = duration_cast<milliseconds>(end - start);

    // now start timing; use the minimum of 10 repetitions for timing. Use the last result.
    Vector timing(repN);
    for(int r=0; r<repN; ++r){
        start = Clock::now();
        mc = mc_nonvec([](double x) {return std::sin(x);}, rnorm, M);
        end = Clock::now();
	    diff = duration_cast<milliseconds>(end - start);
        timing[r] = diff.count();
    }
    cout << "mc_nonvec takes " << *std::min_element(timing.begin(), timing.end())
        << " ms, result = " << mc.mu << " +- " << mc.stat << "\n\n";    

    for(int r=0; r<repN; ++r){
        start = Clock::now();
        mc = mc_scalar([](double x) {return std::sin(x);}, rnorm, M);
        end = Clock::now();
	    diff = duration_cast<milliseconds>(end - start);
        timing[r] = diff.count();
    }
    cout << "mc_scalar takes " << *std::min_element(timing.begin(), timing.end())
        << " ms, result = " << mc.mu << " +- " << mc.stat << "\n\n";

    for(int r=0; r<repN; ++r){
        start = Clock::now();
        mc = mc_max_naive(rnorm, M, N);
        end = Clock::now();
	    diff = duration_cast<milliseconds>(end - start);
        timing[r] = diff.count();
    }
    cout << "mc_max_naive takes " << *std::min_element(timing.begin(), timing.end())
        << " ms, result = " << mc.mu << " +- " << mc.stat << "\n\n";

    for(int r=0; r<repN; ++r){
        start = Clock::now();
        mc = mc_max_partial(rnorm, M, N);
        end = Clock::now();
	    diff = duration_cast<milliseconds>(end - start);
        timing[r] = diff.count();
    }
    cout << "mc_max_partial takes " << *std::min_element(timing.begin(), timing.end())
        << " ms, result = " << mc.mu << " +- " << mc.stat << "\n\n";

    for(int r=0; r<repN; ++r){
        start = Clock::now();
        mc = mc_max(rnorm, M, N);
        end = Clock::now();
	    diff = duration_cast<milliseconds>(end - start);
        timing[r] = diff.count();
    }
    cout << "mc_max takes " << *std::min_element(timing.begin(), timing.end())
        << " ms, result = " << mc.mu << " +- " << mc.stat << "\n\n";

    return 0;
}